# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

"""Some helper function for loading and converting image data."""

import math
from typing import Optional

from gi.repository import GdkPixbuf, Gtk, Gdk, GLib
import cairo


def get_surface_for_pixbuf(widget: Gtk.Widget, pixbuf: Optional[GdkPixbuf.Pixbuf])\
        -> Optional[cairo.Surface]:
    """:returns: a cairo surface, if possible"""
    if not pixbuf:
        return None
    scale_factor = widget.get_scale_factor()
    return Gdk.cairo_surface_create_from_pixbuf(
            pixbuf, scale_factor, widget.get_window())


def get_surface_extents(surface):
    """Gives (x, y, width, height) for a surface, scale independent"""

    ctx = cairo.Context(surface)
    x1, y1, x2, y2 = ctx.clip_extents()
    x1 = int(math.floor(x1))
    y1 = int(math.floor(y1))
    x2 = int(math.ceil(x2))
    y2 = int(math.ceil(y2))
    x2 -= x1
    y2 -= y1

    return (x1, y1, x2, y2)


def get_border_radius(_widgets=[]):
    """Returns the border radius commonly used in the current theme.
    If there are no rounded corners 0 will be returned.
    """

    if not _widgets:
        b = Gtk.Button()
        b.show()
        e = Gtk.Entry()
        e.show()
        _widgets += [b, e]

    radii = []
    for widget in _widgets:
        style_context = widget.get_style_context()
        radii.append(style_context.get_property(
            Gtk.STYLE_PROPERTY_BORDER_RADIUS, style_context.get_state()))
    radius = max(radii)

    # Doesn't work on the default Ubuntu theme.
    # Not sure why, so fix manually for now
    theme_name = Gtk.Settings.get_default().props.gtk_theme_name
    if theme_name in ("Ambiance", "Radiance"):
        radius = int(radius / 1.5)

    return radius


def add_border(pixbuf, color, width=1, radius=0):
    """Add a border to the pixbuf and round of the edges.
    color is a Gdk.RGBA
    The resulting pixbuf will be width * 2px higher and wider.

    Can not fail.
    """

    pi = math.pi

    def new_border_path(ctx, w, h, r, margin=0):
        d = r + margin
        ctx.new_path()
        ctx.arc(w - d, d, r, -pi / 2, 0)
        ctx.arc(w - d, h - d, r, 0, pi / 2)
        ctx.arc(d, h - d, r, pi / 2, pi)
        ctx.arc(d, d, r, pi, pi * 3 / 2)
        ctx.close_path()

    w = pixbuf.get_width() + width * 2
    h = pixbuf.get_height() + width * 2
    r = min(radius, min(w, h) / 2)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
    ctx = cairo.Context(surface)

    new_border_path(ctx, w, h, r + width)
    ctx.clip()
    ctx.set_source_rgba(color.red, color.green, color.blue, color.alpha)
    ctx.paint()

    new_border_path(ctx, w, h, r, width)
    Gdk.cairo_set_source_pixbuf(ctx, pixbuf, width, width)
    ctx.fill()

    return Gdk.pixbuf_get_from_surface(surface, 0, 0, w, h)


def add_border_widget(pixbuf, widget):
    """Like add_border() but uses the widget to get a border color and a
    border width.
    """

    context = widget.get_style_context()
    color = context.get_color(context.get_state())
    color.alpha *= 0.1
    scale_factor = widget.get_scale_factor()
    border_radius = get_border_radius() * scale_factor

    return add_border(pixbuf, color, width=scale_factor, radius=border_radius)


def scale(pixbuf, boundary, scale_up=True, force_copy=False):
    """Scale a pixbuf so it fits into the boundary.
    (preserves image aspect ratio)

    If `scale_up` is True, the resulting pixbuf can be larger than
    the original one.

    If `force_copy` is False the resulting pixbuf might be the passed one.

    Can not fail.
    """

    size = pixbuf.get_width(), pixbuf.get_height()

    scale_w, scale_h = calc_scale_size(boundary, size, scale_up)

    if (scale_w, scale_h) == size:
        if force_copy:
            return pixbuf.copy()
        return pixbuf

    return pixbuf.scale_simple(scale_w, scale_h, GdkPixbuf.InterpType.BILINEAR)


def calc_scale_size(boundary, size, scale_up=True):
    """Returns the biggest possible size to fit into the boundary,
    respecting the aspect ratio.

    If `scale_up` is True the result can be larger than size.

    All sizes have to be > 0.
    """

    bwidth, bheight = boundary
    iwidth, iheight = size

    if bwidth <= 0 or bheight <= 0 or iwidth <= 0 or iheight <= 0:
        raise ValueError

    scale_w, scale_h = iwidth, iheight

    if iwidth > bwidth or iheight > bheight or scale_up:
        bratio = float(bwidth) / bheight
        iratio = float(iwidth) / iheight

        if iratio > bratio:
            scale_w = bwidth
            scale_h = int(bwidth / iratio)
        else:
            scale_w = int(bheight * iratio)
            scale_h = bheight

    return scale_w, scale_h


def pixbuf_from_file(fileobj, boundary, scale_factor=1):
    """Returns a pixbuf with the maximum size defined by boundary.

    Can raise GLib.GError and return None
    """

    try:
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(fileobj.name)
    except GLib.GError:
        try:
            loader = GdkPixbuf.PixbufLoader()
            loader.write(fileobj.read())
            loader.close()
            fileobj.seek(0, 0)
            pixbuf = loader.get_pixbuf()
        except EnvironmentError:
            return

    w, h = boundary
    w *= scale_factor
    h *= scale_factor

    return scale(pixbuf, (w, h), scale_up=True)



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from ._pattern import (Pattern, FileFromPattern, XMLFromPattern,
    XMLFromMarkupPattern, error,
    ArbitraryExtensionFileFromPattern, URLFromPattern)


URLFromPattern
ArbitraryExtensionFileFromPattern
FileFromPattern
Pattern
XMLFromMarkupPattern
XMLFromPattern
error



# Copyright 2015    Christoph Reiter
#           2016-22 Nick Boultbee
#           2019    Peter Strulo
#           2022    Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

from quodlibet import _
from quodlibet import config
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk import Icons
from quodlibet.util.string import decode
from quodlibet.plugins.events import EventPlugin


def _config(section, option, label, tooltip=None, getter=None):
    def on_changed(entry, *args):
        config.settext(section, option, entry.get_text())

    entry = UndoEntry()
    if tooltip:
        entry.set_tooltip_text(tooltip)
    entry.set_text(config.gettext(section, option))
    entry.connect("changed", on_changed)

    def on_reverted(*args):
        config.reset(section, option)
        entry.set_text(config.gettext(section, option))

    revert = Gtk.Button()
    revert.add(Gtk.Image.new_from_icon_name(
        Icons.DOCUMENT_REVERT, Gtk.IconSize.BUTTON))
    revert.connect("clicked", on_reverted)

    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(entry)

    return (lbl, entry, revert)


def text_config(section, option, label, tooltip=None):
    def getter(section, option):
        return decode(config.get(section, option))

    return _config(section, option, label, tooltip, getter)


def boolean_config(section, option, label, tooltip):
    def on_reverted(*args):
        config.reset(section, option)
        button.set_active(config.getboolean(section, option))

    def __toggled(self, section, option):
        config.set(section, option, str(bool(self.get_active())).lower())

    default = config.getboolean(section, option)
    button = Gtk.CheckButton()
    button.set_active(config.getboolean(section, option, default))
    button.set_tooltip_text(tooltip)
    button.connect('toggled', __toggled, section, option)
    revert = Gtk.Button()
    revert.add(Gtk.Image.new_from_icon_name(Icons.DOCUMENT_REVERT, Gtk.IconSize.BUTTON))
    revert.connect("clicked", on_reverted)

    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(button)
    return lbl, button, revert


def int_config(section, option, label, tooltip):
    def getter(section, option):
        return str(config.getint(section, option))

    return _config(section, option, label, tooltip, getter)


def slider_config(section, option, label, tooltip, lower=0, upper=1,
                 on_change_callback=None, label_value_callback=None):
    def on_reverted(*args):
        config.reset(section, option)
        scale.set_value(config.getfloat(section, option))

    def on_change(scale):
        value = scale.get_value()
        if on_change_callback:
            value = on_change_callback(value)
        scale.set_value(value)
        config.set(section, option, value)

    default = config.getfloat(section, option)

    scale = Gtk.HScale.new(Gtk.Adjustment(
                                       value=default,
                                       lower=lower, upper=upper))
    scale.set_value_pos(Gtk.PositionType.LEFT)
    scale.set_show_fill_level(True)
    scale.set_tooltip_text(_(tooltip))

    if label_value_callback:
        scale.connect('format-value', lambda _, value: label_value_callback(value))
    scale.connect('value-changed', on_change)

    revert = Gtk.Button()
    revert.add(Gtk.Image.new_from_icon_name(Icons.DOCUMENT_REVERT, Gtk.IconSize.BUTTON))
    revert.connect("clicked", on_reverted)

    lbl = Gtk.Label(label=label, use_underline=True)
    lbl.set_mnemonic_widget(scale)
    return lbl, scale, revert


class AdvancedPreferences(EventPlugin):
    PLUGIN_ID = "Advanced Preferences"
    PLUGIN_NAME = _("Advanced Preferences")
    PLUGIN_DESC = _("Allow editing of advanced config settings.")
    PLUGIN_CAN_ENABLE = False
    PLUGIN_ICON = Icons.PREFERENCES_SYSTEM

    def __init_defaults(self):
        self.__enabled = False

    def PluginPreferences(self, *args):
        def changed(entry, name, section="settings"):
            config.set(section, name, entry.get_text())

        vb = Gtk.VBox(spacing=12)

        # Tabulate all settings for neatness
        table = Gtk.Table(n_rows=14, n_columns=4)
        table.set_col_spacings(12)
        table.set_row_spacings(6)
        # We don't use translations as these things are internal
        # and don't want to burden the translators...
        # TODO: rethink translation here? (#3494)
        rows = [
            text_config(
                "editing", "id3encoding",
                "ID3 encodings:",
                ("ID3 encodings separated by spaces. "
                 "UTF-8 is always tried first, and Latin-1 is always tried last.")),
            text_config(
                "settings", "search_tags",
                "Extra search tags:",
                ("Tags that get searched in addition to "
                 "the ones present in the song list. Separate with \",\"")),
            text_config("editing", "multi_line_tags",
                        "Multi-line tags:",
                        ("Tags to consider as multi-line (delimited by \\n) "
                         "rather than multi-valued (comma-separated)")),
            text_config("settings", "rating_symbol_full", "Rating symbol (full):"),
            text_config("settings", "rating_symbol_blank", "Rating symbol (blank):"),
            text_config(
                "player", "backend",
                "Backend:",
                "Identifier of the playback backend to use"),
            boolean_config(
                "settings", "disable_hints",
                "Disable hints:",
                "Disable popup windows (treeview hints)"),
            int_config(
                "browsers", "cover_size",
                "Album cover size:",
                ("Size of the album cover images in the album list browser "
                 "(restart required)")),
            boolean_config(
                "settings", "disable_mmkeys",
                "Disable multimedia keys:",
                "(restart required)"),
            text_config(
                "settings", "window_title_pattern",
                "Main window title:",
                ("A (tied) tag for the main window title, e.g. ~title~~people "
                 "(restart required)")),
            text_config(
                "settings", "datecolumn_timestamp_format",
                "Timestamp date format:",
                "A timestamp format for dates, e.g. %Y%m%d %X (restart required)"),
            boolean_config(
                "settings", "scrollbar_always_visible",
                "Scrollbars always visible:",
                ("Toggles whether the scrollbars on the bottom and side of "
                 "the window always are visible or get hidden when not in use "
                 "(restart required)")),
            boolean_config(
                "settings", "monospace_query",
                "Use monospace font for search input:",
                "Helps readability of code-like queries, but looks less consistent "
                "(restart required)"),
            text_config(
                "settings", "query_font_size",
                "Search input font size:",
                "Size to apply to the search query entry, "
                "in any Pango CSS units, e.g. '100%', '1rem'. (restart required)"),
            boolean_config(
                "settings", "pangocairo_force_fontconfig",
                "Force Use Fontconfig Backend:",
                "It's not the default on win/macOS (restart required)"),
            text_config(
                "browsers", "ignored_characters",
                "Ignored characters:",
                "Characters to ignore in queries"),
            boolean_config(
                "settings", "plugins_window_on_top",
                "Plugin window on top:",
                "Toggles whether the plugin window appears on top of others"),
            int_config(
                "autosave", "queue_interval",
                "Queue autosave interval:",
                ("Longest time between play queue auto-saves, or 0 for disabled. "
                 "(restart required)")),
            int_config(
                "browsers", "searchbar_historic_entries",
                "Number of history entries in the search bar:",
                "8 by default (restart advised)"),
            int_config(
                "browsers", "searchbar_enqueue_limit",
                "Search bar confirmation limit for enqueue:",
                ("Maximal size of the song list that can be enqueued from "
                 "the search bar without confirmation.")),
            slider_config(
                "player", "playcount_minimum_length_proportion",
                "Minimum length proportion to consider a track as played:",
                ("Consider a track played after listening to this proportion of "
                 "its total duration"),
                label_value_callback=lambda value: f"{int(value * 100)}%"),
            text_config(
                "browsers", "missing_title_template",
                "Missing title template string:",
                ("Template for building title of tracks when title tag is missing. "
                 "Tags are allowed, like <~basename> <~dirname> <~format> <~length> "
                 "<~#bitrate>, etc. See tags documentation for details.")
            )
        ]

        for (row, (label, widget, button)) in enumerate(rows):
            label.set_alignment(1.0, 0.5)
            table.attach(label, 0, 1, row, row + 1, xoptions=Gtk.AttachOptions.FILL)
            if isinstance(widget, Gtk.CheckButton):
                xoptions = Gtk.AttachOptions.FILL
                widget.set_alignment(0.0, 0.5)
                table.attach(widget, 1, 2, row, row + 1, xoptions=xoptions)
                # This stops checkbox from expanding too big, or shrinking text entries
                blank = Gtk.Label()
                table.attach(blank, 2, 3, row, row + 1,
                             xoptions=Gtk.AttachOptions.EXPAND)
            else:
                xoptions = Gtk.AttachOptions.FILL
                table.attach(widget, 1, 3, row, row + 1, xoptions=xoptions)
            table.attach(button, 3, 4, row, row + 1,
                         xoptions=Gtk.AttachOptions.SHRINK)

        def on_click(button):
            button.hide()
            table.set_no_show_all(False)
            table.show_all()

        button = Gtk.Button(label=_("I know what I'm doing"), use_underline=True)
        button.connect("clicked", on_click)
        vb.pack_start(button, True, True, 0)
        vb.pack_start(table, True, True, 0)
        table.set_no_show_all(True)

        return vb



# Copyright 2013 Simonas Kazlauskas
#           2018 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from os import path

from quodlibet import _
from quodlibet.plugins.cover import cover_dir
from quodlibet.util import print_d
from quodlibet.util.cover.http import ApiCoverSourcePlugin, escape_query_value
from quodlibet.util.path import escape_filename


class LastFMCover(ApiCoverSourcePlugin):
    PLUGIN_ID = "lastfm-cover"
    PLUGIN_NAME = _("Last.fm Cover Source")
    PLUGIN_DESC = _("Downloads covers from Last.fm's cover art archive.")

    @classmethod
    def group_by(cls, song):
        return song.album_key

    @staticmethod
    def priority():
        return 0.33  # No cover size guarantee, accurate

    @property
    def cover_path(self):
        mbid = self.song.get('musicbrainz_albumid', None)
        # It is beneficial to use mbid for cover names.
        if mbid:
            return path.join(cover_dir, escape_filename(mbid))
        else:
            return super().cover_path

    @property
    def url(self):
        _url = 'https://ws.audioscrobbler.com/2.0?method=album.getinfo&' + \
               'api_key=107db6fd4c1c7f53b1526fafddab2c82&format=json&' + \
               'artist={artist}&album={album}&mbid={mbid}'
        song = self.song
        # This can work well for albums in Last.FM
        artists = self._album_artists_for(song) or 'Various Artists'
        song = self.song
        artist = escape_query_value(artists)
        album = escape_query_value(song.get('album', ''))
        mbid = escape_query_value(song.get('musicbrainz_albumid', ''))
        if (artist and album) or mbid:
            return _url.format(artist=artist, album=album, mbid=mbid)
        else:
            return None  # Not enough data

    def _handle_search_response(self, message, json, data=None):
        if not json:
            print_d('Server did not return valid JSON')
            return self.emit('search-complete', [])
        album = json.get('album', {})
        if not album:
            print_d('Album data is not available')
            return self.emit('search-complete', [])
        results = []
        for img in album['image']:
            if img['size'] in ('mega', 'extralarge'):
                url = img['#text']
                if not url:
                    # Yes sometimes it's there but blank
                    continue
                print_d("Got last.fm image: %s" % img)
                results.append({'artist': album['artist'],
                                'album': album['name'],
                                'cover': url.replace('/300x300', '/500x500'),
                                'dimensions': '500x500'
                                })
                # This one can be massive, and slow
                results.append({'artist': album['artist'],
                                'album': album['name'],
                                'cover': url.replace('/300x300', ''),
                                'dimensions': '(original)'
                                })
                # Prefer the bigger ones
                break
        self.emit('search-complete', results)



# Copyright 2018 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from .gnome import GnomeSessionClient


class XfceSessionClient(GnomeSessionClient):

    DBUS_NAME = 'org.xfce.SessionManager'
    DBUS_OBJECT_PATH = '/org/xfce/SessionManager'
    DBUS_MAIN_INTERFACE = 'org.xfce.Session.Manager'
    DBUS_CLIENT_INTERFACE = 'org.xfce.Session.Client'



# Copyright 2010,2012 Christoph Reiter <reiter.christoph@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import time
import tempfile

import dbus
import dbus.service
from senf import fsn2uri

from quodlibet import app
from quodlibet.order.repeat import RepeatListForever, RepeatSongForever
from quodlibet.util.dbusutils import DBusIntrospectable, DBusProperty
from quodlibet.util.dbusutils import dbus_unicode_validate as unival

from .util import MPRISObject

# TODO: OpenUri, CanXYZ
# Date parsing (util?)


# http://www.mpris.org/2.0/spec/
class MPRIS2(DBusProperty, DBusIntrospectable, MPRISObject):

    BUS_NAME = "org.mpris.MediaPlayer2.quodlibet"
    PATH = "/org/mpris/MediaPlayer2"

    ROOT_IFACE = "org.mpris.MediaPlayer2"

    ROOT_ISPEC = """
<method name="Raise"/>
<method name="Quit"/>"""

    ROOT_PROPS = """
<property name="CanQuit" type="b" access="read"/>
<property name="CanRaise" type="b" access="read"/>
<property name="CanSetFullscreen" type="b" access="read"/>
<property name="HasTrackList" type="b" access="read"/>
<property name="Identity" type="s" access="read"/>
<property name="DesktopEntry" type="s" access="read"/>
<property name="SupportedUriSchemes" type="as" access="read"/>
<property name="SupportedMimeTypes" type="as" access="read"/>"""

    PLAYER_IFACE = "org.mpris.MediaPlayer2.Player"

    PLAYER_ISPEC = """
<method name="Next"/>
<method name="Previous"/>
<method name="Pause"/>
<method name="PlayPause"/>
<method name="Stop"/>
<method name="Play"/>
<method name="Seek">
  <arg direction="in" name="Offset" type="x"/>
</method>
<method name="SetPosition">
  <arg direction="in" name="TrackId" type="o"/>
  <arg direction="in" name="Position" type="x"/>
</method>
<method name="OpenUri">
  <arg direction="in" name="Uri" type="s"/>
</method>
<signal name="Seeked">
  <arg name="Position" type="x"/>
</signal>"""

    PLAYER_PROPS = """
<property name="PlaybackStatus" type="s" access="read"/>
<property name="LoopStatus" type="s" access="readwrite"/>
<property name="Rate" type="d" access="readwrite"/>
<property name="Shuffle" type="b" access="readwrite"/>
<property name="Metadata" type="a{sv}" access="read"/>
<property name="Volume" type="d" access="readwrite"/>
<property name="Position" type="x" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" \
value="false"/>
</property>
<property name="MinimumRate" type="d" access="read"/>
<property name="MaximumRate" type="d" access="read"/>
<property name="CanGoNext" type="b" access="read"/>
<property name="CanGoPrevious" type="b" access="read"/>
<property name="CanPlay" type="b" access="read"/>
<property name="CanPause" type="b" access="read"/>
<property name="CanSeek" type="b" access="read"/>
<property name="CanControl" type="b" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" \
value="false"/>
</property>"""

    def __init__(self):
        DBusIntrospectable.__init__(self)
        DBusProperty.__init__(self)

        self.set_introspection(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_ISPEC)
        self.set_properties(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_PROPS)

        self.set_introspection(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_ISPEC)
        self.set_properties(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_PROPS)

        bus = dbus.SessionBus()
        name = dbus.service.BusName(self.BUS_NAME, bus)
        MPRISObject.__init__(self, bus, self.PATH, name)

        self.__metadata = None
        self.__cover = None
        player_options = app.player_options
        self.__repeat_id = player_options.connect(
            "notify::repeat", self.__repeat_changed)
        self.__random_id = player_options.connect(
            "notify::shuffle", self.__shuffle_changed)
        self.__single_id = player_options.connect(
            "notify::single", self.__single_changed)

        self.__lsig = app.librarian.connect("changed", self.__library_changed)
        self.__vsig = app.player.connect("notify::volume",
                                         self.__volume_changed)
        self.__seek_sig = app.player.connect("seek", self.__seeked)

    def remove_from_connection(self, *arg, **kwargs):
        super().remove_from_connection(*arg, **kwargs)

        player_options = app.player_options
        player_options.disconnect(self.__repeat_id)
        player_options.disconnect(self.__random_id)
        player_options.disconnect(self.__single_id)
        app.librarian.disconnect(self.__lsig)
        app.player.disconnect(self.__vsig)
        app.player.disconnect(self.__seek_sig)

        if self.__cover is not None:
            self.__cover.close()
            self.__cover = None
        self.__invalidate_metadata()

    def __volume_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Volume"])

    def __repeat_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["LoopStatus"])

    def __shuffle_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Shuffle"])

    def __single_changed(self, *args):
        self.emit_properties_changed(self.PLAYER_IFACE, ["LoopStatus"])

    def __seeked(self, player, song, ms):
        self.Seeked(ms * 1000)

    def __library_changed(self, library, songs):
        self.__invalidate_metadata()
        if not songs or app.player.info not in songs:
            return
        self.emit_properties_changed(self.PLAYER_IFACE, ["Metadata"])

    @dbus.service.method(ROOT_IFACE)
    def Raise(self):
        app.present()

    @dbus.service.method(ROOT_IFACE)
    def Quit(self):
        app.quit()

    @dbus.service.signal(PLAYER_IFACE, signature="x")
    def Seeked(self, position):
        pass

    @dbus.service.method(PLAYER_IFACE)
    def Next(self):
        player = app.player
        paused = player.paused
        player.next()
        player.paused = paused

    @dbus.service.method(PLAYER_IFACE)
    def Previous(self):
        player = app.player
        paused = player.paused
        player.previous()
        player.paused = paused

    @dbus.service.method(PLAYER_IFACE)
    def Pause(self):
        app.player.paused = True

    @dbus.service.method(PLAYER_IFACE)
    def Play(self):
        app.player.play()

    @dbus.service.method(PLAYER_IFACE)
    def PlayPause(self):
        app.player.playpause()

    @dbus.service.method(PLAYER_IFACE)
    def Stop(self):
        app.player.stop()

    @dbus.service.method(PLAYER_IFACE, in_signature="x")
    def Seek(self, offset):
        new_pos = app.player.get_position() + offset / 1000
        app.player.seek(new_pos)

    @dbus.service.method(PLAYER_IFACE, in_signature="ox")
    def SetPosition(self, track_id, position):
        if track_id == self.__get_current_track_id():
            app.player.seek(position / 1000)

    def paused(self):
        self.emit_properties_changed(self.PLAYER_IFACE, ["PlaybackStatus"])
    unpaused = paused

    def song_started(self, song):
        self.__invalidate_metadata()

        # so the position in clients gets updated faster
        self.Seeked(0)

        self.emit_properties_changed(self.PLAYER_IFACE,
                                    ["PlaybackStatus", "Metadata"])

    def __get_current_track_id(self):
        path = "/net/sacredchao/QuodLibet"
        if not app.player.info:
            return dbus.ObjectPath(path + "/" + "NoTrack")
        return dbus.ObjectPath(path + "/" + str(id(app.player.info)))

    def __invalidate_metadata(self):
        self.__metadata = None

    def __get_metadata(self):
        if self.__metadata is None:
            self.__metadata = self.__get_metadata_real()
            assert self.__metadata is not None
        return self.__metadata

    def __get_metadata_real(self):
        """
        https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/
        """

        metadata = {}
        metadata["mpris:trackid"] = self.__get_current_track_id()

        def ignore_overflow(dbus_type, value):
            try:
                return dbus_type(value)
            except OverflowError:
                return 0

        song = app.player.info
        if not song:
            return metadata

        metadata["mpris:length"] = ignore_overflow(
            dbus.Int64, song("~#length") * 10 ** 6)

        if self.__cover is not None:
            self.__cover.close()
            self.__cover = None

        cover = app.cover_manager.get_cover(song)
        if cover:
            is_temp = cover.name.startswith(tempfile.gettempdir())
            if is_temp:
                self.__cover = cover
            metadata["mpris:artUrl"] = fsn2uri(cover.name)

        # All list values
        list_val = {"artist": "artist", "albumArtist": "albumartist",
            "comment": "comment", "composer": "composer", "genre": "genre",
            "lyricist": "lyricist"}
        for xesam, tag in list_val.items():
            vals = song.list(tag)
            if vals:
                metadata["xesam:" + xesam] = list(map(unival, vals))

        # All single values
        sing_val = {"album": "album", "title": "title", "asText": "~lyrics"}
        for xesam, tag in sing_val.items():
            vals = song.comma(tag)
            if vals:
                metadata["xesam:" + xesam] = unival(vals)

        # URI
        metadata["xesam:url"] = song("~uri")

        # Integers
        num_val = {"audioBPM": "bpm", "discNumber": "disc",
                   "trackNumber": "track", "useCount": "playcount"}

        for xesam, tag in num_val.items():
            val = song("~#" + tag, None)
            if val is not None:
                metadata["xesam:" + xesam] = ignore_overflow(dbus.Int32, val)

        # Rating
        metadata["xesam:userRating"] = ignore_overflow(
            dbus.Double, song("~#rating"))

        # Dates
        ISO_8601_format = "%Y-%m-%dT%H:%M:%S"
        tuple_time = time.gmtime(song("~#lastplayed"))
        iso_time = time.strftime(ISO_8601_format, tuple_time)
        metadata["xesam:lastUsed"] = iso_time

        year = song("~year")
        if year:
            try:
                tuple_time = time.strptime(year, "%Y")
                iso_time = time.strftime(ISO_8601_format, tuple_time)
            except ValueError:
                pass
            else:
                metadata["xesam:contentCreated"] = iso_time

        return metadata

    def set_property(self, interface, name, value):
        player = app.player
        player_options = app.player_options

        if interface == self.PLAYER_IFACE:
            if name == "LoopStatus":
                if value == "Playlist":
                    player_options.repeat = True
                    app.window.order.repeater = RepeatListForever
                elif value == "Track":
                    player_options.repeat = True
                    app.window.order.repeater = RepeatSongForever
                elif value == "None":
                    player_options.repeat = False
                    app.window.order.repeater = RepeatListForever
            elif name == "Rate":
                pass
            elif name == "Shuffle":
                player_options.shuffle = value
            elif name == "Volume":
                player.volume = value

    def get_property(self, interface, name):
        player = app.player
        player_options = app.player_options

        if interface == self.ROOT_IFACE:
            if name == "CanQuit":
                return True
            elif name == "CanRaise":
                return True
            elif name == "CanSetFullscreen":
                return False
            elif name == "HasTrackList":
                return False
            elif name == "Identity":
                return app.name
            elif name == "DesktopEntry":
                return "io.github.quodlibet.QuodLibet"
            elif name == "SupportedUriSchemes":
                # TODO: enable once OpenUri is done
                can = lambda s: False
                #can = lambda s: app.player.can_play_uri("%s://fake" % s)
                schemes = ["http", "https", "ftp", "file", "mms"]
                return filter(can, schemes)
            elif name == "SupportedMimeTypes":
                from quodlibet import formats
                return formats.mimes
        elif interface == self.PLAYER_IFACE:
            if name == "PlaybackStatus":
                if not player.song:
                    return "Stopped"
                return ("Playing", "Paused")[int(player.paused)]
            elif name == "LoopStatus":
                if not player_options.repeat:
                    return "None"
                else:
                    if player_options.single:
                        return "Track"
                    return "Playlist"
            elif name == "Rate":
                return 1.0
            elif name == "Shuffle":
                return player_options.shuffle
            elif name == "Metadata":
                return self.__get_metadata()
            elif name == "Volume":
                # https://gitlab.freedesktop.org/mpris/mpris-spec/issues/8
                return player.volume
            elif name == "Position":
                return player.get_position() * 1000
            elif name == "MinimumRate":
                return 1.0
            elif name == "MaximumRate":
                return 1.0
            elif name == "CanGoNext":
                return True
            elif name == "CanGoPrevious":
                return True
            elif name == "CanPlay":
                return True
            elif name == "CanPause":
                return True
            elif name == "CanSeek":
                return True
            elif name == "CanControl":
                return True



# Copyright (c) 2018 Philipp Wolfer <ph.wolfer@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from typing import Optional
import json
import logging
import os
import ssl
import time
from http.client import HTTPConnection, HTTPSConnection

HOST_NAME: Optional[str] = "api.listenbrainz.org"
PATH_SUBMIT = "/1/submit-listens"
SSL_CONTEXT: Optional[ssl.SSLContext] = ssl.create_default_context()

# to run against a local dev server
if os.getenv("QL_LISTENBRAINZ_DEV_SERVER") is not None:
    HOST_NAME = os.getenv("QL_LISTENBRAINZ_DEV_SERVER")
    SSL_CONTEXT = None


class Track:
    """
    Represents a single track to submit.

    See https://listenbrainz.readthedocs.io/en/latest/dev/json.html
    """
    def __init__(self, artist_name, track_name,
                 release_name=None, additional_info={}):
        """
        Create a new Track instance
        @param artist_name as str
        @param track_name as str
        @param release_name as str
        @param additional_info as dict
        """
        self.artist_name = artist_name
        self.track_name = track_name
        self.release_name = release_name
        self.additional_info = additional_info

    @staticmethod
    def from_dict(data):
        return Track(
            data["artist_name"],
            data["track_name"],
            data.get("release_name", None),
            data.get("additional_info", {})
        )

    def to_dict(self):
        return {
            "artist_name": self.artist_name,
            "track_name": self.track_name,
            "release_name": self.release_name,
            "additional_info": self.additional_info
        }

    def __repr__(self):
        return "Track(%s, %s)" % (self.artist_name, self.track_name)


class ListenBrainzClient:
    """
    Submit listens to ListenBrainz.org.

    See https://listenbrainz.readthedocs.io/en/latest/dev/api.html
    """

    def __init__(self, logger=logging.getLogger(__name__)):
        self.__next_request_time = 0
        self.user_token = None
        self.logger = logger

    def listen(self, listened_at, track):
        """
        Submit a listen for a track
        @param listened_at as int
        @param entry as Track
        """
        payload = _get_payload(track, listened_at)
        return self._submit("single", [payload])

    def playing_now(self, track):
        """
        Submit a playing now notification for a track
        @param track as Track
        """
        payload = _get_payload(track)
        return self._submit("playing_now", [payload])

    def import_tracks(self, tracks):
        """
        Import a list of tracks as (listened_at, Track) pairs
        @param track as [(int, Track)]
        """
        payload = _get_payload_many(tracks)
        return self._submit("import", payload)

    def _submit(self, listen_type, payload, retry=0):
        self._wait_for_ratelimit()
        self.logger.debug("ListenBrainz %s: %r", listen_type, payload)
        data = {
            "listen_type": listen_type,
            "payload": payload
        }
        headers = {
            "Authorization": "Token %s" % self.user_token,
            "Content-Type": "application/json"
        }
        body = json.dumps(data)
        print("submit: %s" % body)
        if SSL_CONTEXT is not None:
            conn = HTTPSConnection(HOST_NAME, context=SSL_CONTEXT)
        else:
            conn = HTTPConnection(HOST_NAME)
        # XXX TODO, catch errors?
        conn.request("POST", PATH_SUBMIT, body, headers)
        response = conn.getresponse()
        response_text = response.read()
        try:
            response_data = json.loads(response_text)
        #Python3
        #except json.JSONDecodeError:
        #    response_data = response_text
        #Python2
        except ValueError as e:
            if str(e) != "No JSON object could be decoded":
                raise e
            response_data = response_text

        self._handle_ratelimit(response)
        log_msg = "Response %s: %r" % (response.status, response_data)
        if response.status == 429 and retry < 5:  # Too Many Requests
            self.logger.warning(log_msg)
            return self._submit(listen_type, payload, retry + 1)
        elif response.status == 200:
            self.logger.debug(log_msg)
        else:
            self.logger.error(log_msg)
        return response

    def _wait_for_ratelimit(self):
        now = time.time()
        if self.__next_request_time > now:
            delay = self.__next_request_time - now
            self.logger.debug("Rate limit applies, delay %d", delay)
            time.sleep(delay)

    def _handle_ratelimit(self, response):
        remaining = int(response.getheader("X-RateLimit-Remaining", 0))
        reset_in = int(response.getheader("X-RateLimit-Reset-In", 0))
        self.logger.debug("X-RateLimit-Remaining: %i", remaining)
        self.logger.debug("X-RateLimit-Reset-In: %i", reset_in)
        if remaining == 0:
            self.__next_request_time = time.time() + reset_in


def _get_payload_many(tracks):
    payload = []
    for (listened_at, track) in tracks:
        data = _get_payload(track, listened_at)
        payload.append(data)
    return payload


def _get_payload(track, listened_at=None):
    data = {
        "track_metadata": track.to_dict()
    }
    if listened_at is not None:
        data["listened_at"] = listened_at
    return data



# Copyright 2016 Mice Pápai
#           2018 Nick Boultbee
#
# Based on lastfm.py by Simonas Kazlauskas
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


from os import path

import re
from gi.repository import Soup

from quodlibet import _
from quodlibet.plugins.cover import cover_dir
from quodlibet.util.http import download_json
from quodlibet.util.cover.http import escape_query_value
from quodlibet.util.cover.http import ApiCoverSourcePlugin
from quodlibet.util.path import escape_filename
from quodlibet.util import print_d


class DiscogsCover(ApiCoverSourcePlugin):
    PLUGIN_ID = "discogs-cover"
    PLUGIN_NAME = _("Discogs Cover Source")
    PLUGIN_DESC = _("Downloads covers from Discogs.")

    credentials = ('key=aWfZGjHQvkMcreUECGAp' +
                   '&secret=VlORkklpdvAwJMwxUjNNSgqicjuizJAl')
    use_secondary = True

    @classmethod
    def group_by(cls, song):
        return song.album_key

    @staticmethod
    def priority():
        return 0.6

    @property
    def cover_path(self):
        mbid = self.song.get('musicbrainz_albumid', None)
        if mbid:
            return path.join(cover_dir, escape_filename(mbid))
        else:
            return super().cover_path

    @property
    def url(self):
        _url = ('https://api.discogs.com/database/search?' +
                self.credentials +
                '&format=CD&per_page=5'
                '&type=release' +
                '&artist={artist}' +
                '&release_title={album}')
        # Discogs seems to use 'Various' almost exclusively for compilations
        artists = self._album_artists_for(self.song) or 'Various'
        if 'various artists' in artists.lower():
            artists = 'Various'
        artist = escape_query_value(artists)
        album = escape_query_value(self.song.get('album', ''))
        if artist and album:
            return _url.format(artist=artist, album=album)
        else:
            return None   # Not enough data

    def _handle_search_response(self, message, json_dict, data=None):
        if not json_dict:
            print_d('Server did not return any valid JSON')
            return self.emit('search-complete', [])
        try:
            results = json_dict.get('results', [])
            covers = filter(None, (self.result_for(r.get('cover_image', None))
                                   for r in results))
            if covers:
                return self.emit('search-complete', list(covers))
            res_url = results[0].get('resource_url', '')
        except IndexError:
            return self.emit('search-complete', [])
        else:
            msg = Soup.Message.new('GET',
                                   "%s?%s" % (res_url, self.credentials))
            download_json(msg, self.cancellable, self._handle_album_data, None)

    def _handle_album_data(self, message, json_dict, data=None):

        images = json_dict.get('images', None)

        if not images:
            print_d('Covers are not available')
            return self.emit('search-complete', [])

        results = list(self._covers_of_type(images))
        if not results and images and self.use_secondary:
            results.append(next(self._covers_of_type(images, "secondary")))

        self.emit('search-complete', results)

    def _covers_of_type(self, images, image_type='primary'):
        for image in images:
            url = image.get('uri')
            if url and image['type'] == image_type:
                yield self.result_for(url)

    def result_for(self, url):
        if not url:
            return None
        dimensions = re.compile(r'/(\d+x\d+)/').search(url)
        dimensions = dimensions and dimensions.group(1)
        if dimensions:
            dims = map(int, dimensions.split('x'))
            if min(dims) < self.MIN_DIMENSION:
                print_d("%s is too small to use" % dimensions)
                return None
        return {'cover': url, 'dimensions': dimensions}
